import os
from contextlib import contextmanager
from sqlalchemy import MetaData
from sqlalchemy.engine import Engine, create_engine
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.ext.declarative import DeclarativeMeta
from sqlalchemy.orm import sessionmaker, Session, declarative_base
from typing import Generator, Optional, Type
from maleo_foundation.types import BaseTypes
from maleo_foundation.utils.logger import BaseLogger
from maleo_foundation.db.table import BaseTable

class MetadataManager:
    Base:DeclarativeMeta = declarative_base()
    metadata:MetaData = Base.metadata

class SessionManager:
    def __init__(self, logger:BaseLogger, engine:Engine):
        self._logger = logger
        self._logger.info("Initializing SessionMaker")
        self._sessionmaker:sessionmaker[Session] = sessionmaker(bind=engine, expire_on_commit=False)
        self._logger.info("SessionMaker initialized successfully")

    def _session_handler(self) -> Generator[Session, None, None]:
        """Reusable function for managing database sessions."""
        session = self._sessionmaker()
        self._logger.debug("New database session created.")
        try:
            yield session  #* Provide session
            session.commit()  #* Auto-commit on success
        except SQLAlchemyError as e:
            session.rollback()  #* Rollback on error
            self._logger.error(f"[SQLAlchemyError] Database transaction failed: {e}", exc_info=True)
            raise
        except Exception as e:
            session.rollback()  #* Rollback on error
            self._logger.error(f"[Exception] Database transaction failed: {e}", exc_info=True)
            raise
        finally:
            session.close()  #* Ensure session closes
            self._logger.debug("Database session closed.")

    def inject(self) -> Generator[Session, None, None]:
        """Returns a generator that yields a SQLAlchemy session for dependency injection."""
        return self._session_handler()

    @contextmanager
    def get(self) -> Generator[Session, None, None]:
        """Context manager for manual session handling. Supports `with SessionManager.get() as session:`"""
        yield from self._session_handler()

    def dispose(self) -> None:
        """Dispose of the sessionmaker and release any resources."""
        if self._sessionmaker is not None:
            self._sessionmaker.close_all()
            self._sessionmaker = None

        self._logger.info("SessionManager disposed successfully")
        self._logger = None

class DatabaseManager:
    def __init__(
        self,
        metadata:MetaData,
        logger:BaseLogger,
        url:BaseTypes.OptionalString = None
    ):
        self._metadata = metadata #* Define database metadata
        self._logger = logger #* Define database logger

        #* Create engine
        url = url or os.getenv("DB_URL")
        if url is None:
            raise ValueError("DB_URL environment variable must be set if url is not provided")
        self._logger.info("Creating SQlAlchemy engine")
        self._engine = create_engine(url=url, echo=False, pool_pre_ping=True, pool_recycle=3600)
        self._logger.info("SQlAlchemy engine created successfully")

        #* Creating all table from metadata
        self._logger.info("Creating all tables defined in metadata")
        self._metadata.create_all(bind=self._engine) #* Create all tables
        self._logger.info("Created all tables defined in metadata")

        #* Initializing session manager
        self._logger.info("Initializing session manager")
        self._session = SessionManager(logger=self._logger, engine=self._engine) #* Define session
        self._logger.info("Session manager initialized successfully")

    @property
    def metadata(self) -> MetaData:
        return self._metadata

    @property
    def engine(self) -> Engine:
        return self._engine

    @property
    def session(self) -> SessionManager:
        return self._session

    def dispose(self) -> None:
        #* Dispose session
        if self._session is not None:
            self._logger.info("Disposing Session Manager")
            self._session.dispose()
            self._session = None
        #* Dispose engine
        if self._engine is not None:
            self._logger.info("Disposing Engine Manager")
            self._engine.dispose()
            self._engine = None
            self._logger.info("Engine Manager disposed successfully")
        #* Dispose metadata
        if self._metadata is not None:
            self._logger.info("Disposing DB Metadata")
            self._metadata = None
            self._logger.info("DB Metadata diposed succesfully")